with Lower_Layer_UDP;
with Ada.Streams;
with Ada.Exceptions;
with Interfaces.C;
with Ada.Numerics.Discrete_Random;
with Ada.Streams.Stream_IO;

package body STCP is

   package C renames Interfaces.C;

   function htonl (Data: in C.unsigned_long) return C.unsigned_long;
   pragma Import (C, Htonl, "htonl");

   function ntohl (Data: in C.unsigned_long) return C.unsigned_long;
   pragma Import (C, Ntohl, "ntohl");

   function H_To_N_Int_32 (N: Int_32) return Int_32 is
   begin
      return Int_32 (htonl (C.Unsigned_Long (N)));
   end H_To_N_Int_32;

   function N_To_H_Int_32 (N: Int_32) return Int_32 is
   begin
      return Int_32 (ntohl (C.Unsigned_Long (N)));
   end N_To_H_Int_32;



   package Int_32_Random is new Ada.Numerics.Discrete_Random (Int_32);
   G: Int_32_Random.Generator;
   package Int_Random is new Ada.Numerics.Discrete_Random (Integer);
   G2: Int_Random.Generator;

   type STCP_Type is (DATA, ACK, SYN, FIN, RST);

   Default_Timeout: constant Duration := 1.0;
   Null_Window: constant Int_32 := 0;
   Null_Data: Ada.Streams.Stream_Element_Array (1..0);


   ProbFallo: Integer:= 0;
      -- Variables para controlar el estado de la conexin
   Enviados : Int_32:= 0;
   Asentidos: Int_32:= 0;
   Almacenados: Int_32:= 0;
   Ventana: Int_32 := 0;
   VentanaEf: Int_32 := 0;
   Terminar: Boolean :=False;

   Open_Connections :
     array (1..Max_Open_Connections) of aliased STCP_Connection_Data;


   -- *******************************************************************
   -- *                            SEND_PACKET                          *
   ---* --------------------------------------------------------------- *
   -- *   Manda un mensaje al receptor, ya sea de datos, syn ...        *
   -- *   nicamente manda el mensaje, no espera respuesta              *
   -- *******************************************************************
   procedure Send_Packet (Sender_EP: LLU.End_Point_Type;
                          Receiver_EP: LLU.End_Point_Type;
                          P_Type: STCP_Type;
                          P_Window: Int_32;
                          P_Seq_Number: Int_32;
                          P_Raw_Data: Ada.Streams.Stream_Element_Array :=
                            Null_Data) is

      Buffer: aliased LLU.Buffer_Type (1024);
      Expected_Ack: Int_32;
      Num: Integer;
   begin

      Num := ((Int_Random.Random(G2)) mod 100)+1;
      --Put_Line ("ALEATORIO:" & Integer'Image(Num));
      if (num > ProbFallo) then
         Put ("Manda paquete. NSeq: " & Int_32'Image(P_Seq_Number));
         if P_Type = SYN or P_Type = FIN then
            Expected_Ack := P_Seq_Number;
         elsif P_Type = DATA then
            Expected_Ack := (P_Seq_Number + 1) mod Int_32'Last;
         end if;

         LLU.Reset (Buffer);
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (STCP_Type'Pos (P_Type) + 1));
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (P_Window));
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (P_Seq_Number));
         Ada.Streams.Stream_Element_Array'Write (Buffer'Access, P_Raw_Data);
         LLU.Send_From (Sender_EP, Receiver_EP, Buffer'Access);
      else
         Put ("Pierde paquete. NSeq: " & Int_32'Image(P_Seq_Number));
      end if;
   end Send_Packet;



   -- *******************************************************************
   -- *                           OPEN                                  *
   -- * --------------------------------------------------------------- *
   -- *  Funcin utilizada por el sender. Manda un mensaje de syn       *
   -- *  y repite hasta 3 veces si no obtiene respuesta                 *
   -- *******************************************************************
   function Open (Sending_EP   : Lower_Layer_UDP.End_Point_Type;
                  Receiving_EP : Lower_Layer_UDP.End_Point_Type;
                  Prob : integer)
                  return STCP_Connection is

      I: Natural := 1;
      Buffer: aliased LLU.Buffer_Type (1024);
      Expired: Boolean;
      Recv_Type: STCP_Type;
      Recv_Win: Int_32;
      Recv_Ack: Int_32;
      Expected_Ack: Int_32;
      N_Veces: Integer:= 0;
   begin

      ProbFallo := Prob;
      while Open_Connections(I).Used loop
         I := I + 1;
      end loop;
      Open_Connections(I).Used := True;
      Open_Connections(I).Sender_EP := Sending_EP;
      Open_Connections(I).Receiver_EP := Receiving_EP;
      Open_Connections(I).Next_Seq_Number := Int_32_Random.Random (G);

      Expected_Ack := Open_Connections(I).Next_Seq_Number;

      -- REPITE TRES VECES ESTE BUCLE SI NO RESPONDE EL SERVIDOR

      loop
         Put_Line ("Manda SYN con Nsec: " &
                   Int_32'Image (Open_Connections(I).Next_Seq_Number));
         LLU.Reset (Buffer);
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (STCP_Type'Pos (SYN) + 1));
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (Null_Window));
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (Open_Connections(I).Next_Seq_Number));
         Ada.Streams.Stream_Element_Array'Write (Buffer'Access, Null_Data);
         LLU.Send_From (Open_Connections(I).Sender_EP,
                        Open_Connections(I).Receiver_EP, Buffer'Access);
         N_Veces := N_Veces +1;
         LLU.Receive (Open_Connections(I).Sender_EP, Buffer'Access,
                      Default_Timeout, Expired);
         if not Expired then
            Recv_Type :=
              STCP_Type'Val (N_To_H_Int_32 (Int_32'Input (Buffer'Access))-1);
            Recv_Win := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
            Recv_Ack := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
         end if;
         exit when
           (not Expired and
            Recv_Type = ACK and
            Recv_Ack = Expected_Ack) or (N_Veces=3);
      end loop;

      -- SI HA RESPONDIDO EL SERVIDOR

      if (not Expired) and (Recv_Type = ACK and Recv_Ack=Expected_Ack) then
         Ventana := Recv_Win;
         Ventanaef:= Ventana;
         Tarea.Start ( I);
         Put ("Recibo ACK: " & Int_32'Image (Recv_Ack));
         Put_Line ("  w:" & Int_32'Image(Ventana));

      else -- SI NO HA RESPONDIDO EL SERVIDOR

         Open_Connections(I).Used:=False;
         Put_Line ("ERROR: Servidor no responde, mando RESET");
         LLU.Reset (Buffer);
         Int_32'Output (Buffer'Access,
                        H_To_N_Int_32 (Stcp_Type'Pos(RST) +1));
         Int_32'Output (Buffer'Access, H_To_N_Int_32 (Null_Window));
         Int_32'Output (Buffer'Access,
                           H_To_N_Int_32 (Open_Connections(I).Next_Seq_Number));
         AS.Stream_Element_Array'Write (Buffer'Access, Null_Data);
         LLU.Send_From (Open_Connections(I).Sender_Ep,
                        Open_Connections(I).Receiver_Ep, Buffer'Access);
      end if;

      return Open_Connections(I)'Access;
   end Open;


   -- ******************************************************************
   -- *                              VALIDA                            *
   -- * -------------------------------------------------------------- *
   -- *   Para comprobar si una conexin se ha abierto satisfactoriam. *
   -- ******************************************************************
   function Valida (Connection: STCP_Connection) return Boolean is
   begin
      return (Connection.Used);
   end Valida;


   -- *****************************************************************
   -- *                            SEND                               *
   -- * ------------------------------------------------------------- *
   -- *   Procedimiento utilizado por el sender para enviar datos.    *
   -- *   En realidad aqu solo se almacenan los datos en una lista   *
   -- *****************************************************************
   procedure Send (Connection : STCP_Connection;
                   Raw_Data   : Ada.Streams.Stream_Element_Array) is

      Elem: Lista.Element;
      Last: AS.Stream_Element_Offset;
   begin

      -- RELLENA LOS DATOS DEL ELEMENTO DE DATOS

      Elem.NSec := Connection.Next_Seq_Number;
      Put ("Meto en lista con Nsec: " & Int_32'Image(Connection.Next_Seq_Number));
      last := (Raw_Data'length);
      Put_Line (AS.Stream_Element_Offset'Image (Last));
      Elem.Paq(1..Last) := Raw_Data(1..last);
      Elem.Len := Last;

      -- Y LO INSERTA EN LA LISTA DE PAQUETES

      L.Insert_Tail (ListaPaq, Elem);
      Almacenados := Almacenados +1;
      Connection.Next_Seq_Number :=
        (Connection.Next_Seq_Number + 1) mod Int_32'Last;
   end Send;


   -- ****************************************************************
   -- *                            CLOSE                             *
   -- * ------------------------------------------------------------ *
   -- *   Procedimiento utilizado por el sender para cerrar la       *
   -- *   conexin. En realidad aqu no se cierra, sino que se       *
   -- *   indica que no hay ms datos, y que hay que cerrarla        *
   -- ****************************************************************
   procedure Close (Connection: STCP_Connection) is
   begin
      Terminar:=True;
      --Put_Line ("No hay ms paquetes. Debo cerrar conexin");
      Tarea.Stop;
   end Close;

   -- ****************************************************************
   -- *                           CLOSECON                           *
   -- * ------------------------------------------------------------ *
   -- *   Realmente aqu se cierra la conexin. Se transmite un fin  *
   -- *   hasta que es asentido por el receptor                      *
   -- ****************************************************************
   procedure CloseCon (Send, Recv: LLU.End_Point_Type; Seq_Number: Int_32) is

      Buffer: aliased LLU.Buffer_Type(1024);
      Expired: Boolean;
      Recv_Type: STCP_Type;
      Recv_Win: Int_32;
      Recv_Ack: Int_32;
   begin

      -- REPITE ESTE BUCLE HASTA QUE RESPONDE EL SERVIDOR AL FIN
      loop
         Put_Line ("CLOSE:Mando FIN con Nsec: " & Int_32'Image(Seq_Number));
         Send_Packet(Send,
                     recv,
                     FIN,
                     Null_Window,
                     Seq_Number,
                     Null_Data);
         Put_Line("");
         LLU.Receive (send, Buffer'Access,
                      Default_Timeout, Expired);
         if not Expired then
            Recv_Type :=
              STCP_Type'Val (N_To_H_Int_32 (Int_32'Input (Buffer'Access))-1);
            Recv_Win := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
            Recv_Ack := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
            Put_Line ("CLOSE: Recibo ack: " & Int_32'Image(Recv_Ack));
         end if;
         exit when
           not Expired and
           Recv_Type = ACK and
           Recv_Ack = Seq_Number;
      end loop;
   end Closecon;


   -- *****************************************************************
   -- *                       TAREA: RecAck                           *
   -- * ------------------------------------------------------------- *
   -- *  En realidad es esta tarea la que toma todas las decisiones   *
   -- *  sobre que paquetes se deben mandar y la que recibe tambin   *
   -- *  todos los acks                                               *
   -- *****************************************************************

   task body RecAck is

      I: Integer;
      X,Y: Float;
      Buffer: aliased LLU.Buffer_Type (1024);
      Expired: Boolean := false;
      Recv_Type: STCP_Type;
      Recv_Win: Int_32;
      Recv_Ack: Int_32;
      Expected_Ack: Int_32;
      N_Veces: Integer;

      Ref: L.Reference;
      Elem: Lista.Element;
      Aux1, Aux2: Int_32;
      Seq_Number: Int_32;
      TerminarTarea : Boolean := False;
      NumEnviados: Int_32 := 0;

      -- Para SlowStart
      CongestionWindow: Int_32:=1;
      MaxWin: Int_32;
      MedWin: Int_32 := 0;

      -- Para Retransmisin/Recuperacin rpida
      UltimoACK : Int_32 := 0;
      NVecesACK: Int_32 := 0;

   begin
      while (not TerminarTarea) loop
         select

            -- COMIENZA LA TAREA
            accept Start (Num:Integer) do
               I:= Num;
            end start;

            -- REPITE ESTE BUCLE HASTA QUE HAYA HECHO UN CLOSE EL SENDER
            -- Y SE HAYAN ASENTIDO TOTOS LOS PAQUETES

            while (not Terminar) or (not(Asentidos=Almacenados)) loop

               -- SE MANDAN LOS PAQUETES QUE SE PUEDA

               if (Asentidos<Almacenados) then
                  Ventanaef := Ventana;
                  Numenviados:=0;
                  -- Avanzo en la lista hasta llegar al primer elemento
                  -- que debo mandar
                  Ref := L.First(ListaPaq);
                  for aux1 in 1..Asentidos loop
                     L.Forward(ListaPaq, Ref);
                     NumEnviados:= NumEnviados +1;
                  end loop;

                  -- Calculo el nmero de elementos que puedo mandar
                  if (CongestionWindow < Ventana) then
                     MaxWin := CongestionWindow;
                  else
                     Maxwin := Ventana;
                  end if;
                  if (Almacenados-Asentidos)>MaxWin then
                     Aux1:= MaxWin;
                  else
                     Aux1:= Almacenados-Asentidos;
                  end if;

                  -- MANDO TODOS LOS ELEMENTOS QUE HE CALCULADO

                  for aux2 in 1..aux1 loop
                     if (not (L.Is_Out(ListaPaq,Ref))) then
                        Elem := L.Get_Element (ListaPaq, Ref);


                        Send_Packet(Open_Connections(I).Sender_EP,
                                    Open_Connections(I).Receiver_EP,
                                    DATA,
                                    Null_Window,
                                    Elem.Nsec,
                                    Elem.Paq(1..Elem.Len));
                        --Put ("Mando paq con nseq: " & Int_32'Image(Elem.NSec));
                        Put ("  w: " & Int_32'Image(Ventana));
                        Put_Line (" wEf: " & Int_32'Image(Ventanaef));

                        Ventanaef := Ventanaef -1;
                        NumEnviados := NumEnviados +1;

                        L.Forward(Listapaq, Ref);
                     end if;
                  end loop;
                  if Numenviados > Enviados then
                     Enviados := NumEnviados;
                  end if;
               end if;

               -- RECIBO TODOS LOS ACK QUE ENVIA EL SERVIDOR

               loop
                  LLU.Receive (Open_Connections(I).Sender_EP, Buffer'Access,
                               Default_Timeout, Expired);
                  if not Expired then -- Si recibo algo
                     Recv_Type :=
                       STCP_Type'Val (N_To_H_Int_32 (Int_32'Input (Buffer'Access))-1);
                     Recv_Win := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
                     Recv_Ack := N_To_H_Int_32 (Int_32'Input (Buffer'Access));
                     if Recv_Type = ACK then
                        Put_Line ("Recibo ACK: " & Int_32'Image(Recv_Ack));
                        -- Actualizo variables
                        Ventana := Recv_Win;
                        Ventanaef := ventana;
                        -- Calculo el el nmero del elemento
                        -- recibido para actualizar variables
                        Ref:= L.First(ListaPaq);
                        Aux1 := 1;
                        --Put_Line ("Primero lista: " & Int_32'image(L.Get_Element(ListaPaq,Ref).Nsec));
                        while ((L.Get_Element(ListaPaq, Ref).NSec) /= Recv_Ack-1) loop
                           L.Forward (ListaPaq, Ref);
                           Aux1 := Aux1+1;
                        end loop;
                        if (Asentidos<Aux1) then
                           Asentidos := Aux1;
                        end if;
                        Put_Line ("Asentidos= " & Int_32'Image(Asentidos));
                        if Asentidos=Almacenados then -- Para enviar el fin
                           Seq_Number := Recv_Ack;
                        end if;
                        -- Para retransmisin/recuperacin rpida
                        if Recv_Ack /= Ultimoack then
                           UltimoACK := Recv_Ack;
                           NVecesACK := 1;
                        else
                           NVecesACK:= NVecesACK +1;
                        end if;

                        if NvecesACK = 4 then -- retransmito
                           Ref := L.First(ListaPaq);

                           while ((L.Get_Element(ListaPaq, Ref).NSec) /= Recv_Ack) loop
                              L.Forward (ListaPaq, Ref);
                              Aux1 := Aux1+1;
                           end loop;

                           Elem:= L.Get_Element(ListaPaq, Ref);
                           Put_Line ("++++++++++++++++++++++++++++++++++++++++++++++++");
                           Put_Line ("+++++++ RETRANSMITO paq con nseq: " &
                                     Int_32'Image(Elem.Nsec));
                           Send_Packet(Open_Connections(I).Sender_EP,
                                       Open_Connections(I).Receiver_EP,
                                       DATA,
                                       Null_Window,
                                       Elem.Nsec,
                                       Elem.Paq(1..Elem.Len));
                           Put_Line("");
                           NvecesACK:=0;
                        end if;

                     end if;
                  end if;

                  exit when Expired;
               end loop;

               if MedWin = 0 then
                  if Asentidos<Enviados then
                     MedWin := CongestionWindow /2;
                     CongestionWindow := 1;
                     Put_Line ("*******  EMPIEZA AIMD  *******");
                  else
                     Congestionwindow := Congestionwindow * 2;
                  end if;
               else
                  if CongestionWindow >= Medwin then
                     if Asentidos < Enviados then
                        CongestionWindow := CongestionWindow /2;
                     else
                        CongestionWindow := Congestionwindow +1;
                     end if;
                  else
                     if Asentidos < Enviados then
                        Medwin := Congestionwindow / 2;
                        CongestionWindow :=1;
                     else
                        Congestionwindow := Congestionwindow *2;
                     end if;
                  end if;
               end if;

               -- Si CongestionWindow = 0 empiezo desde el principio
               if CongestionWindow = 0 then
                  Medwin:= 0;
                  Congestionwindow := 1;
               end if;
               -- ESTADSTICAS
               Put_Line ("*********************************************");
               Put ("Alm: " & Int_32'Image(Almacenados));
               Put ("  Env: " & Int_32'Image(Enviados));
               Put ("  Asen: " & Int_32'Image(Asentidos));
               Put ("  ConW: " & Int_32'Image(CongestionWindow));
               Put ("  MedW: " & Int_32'Image(Medwin));
               Put_Line (" Terminar: " & Boolean'Image(Terminar));
               Put_Line ("*********************************************");

            end loop;
            CloseCon(Open_Connections(I).Sender_EP,
                     Open_Connections(I).Receiver_EP,Seq_number);
         or
            accept Stop do
               TerminarTarea:= True;
            end Stop;
         end select;
      end loop;
   end RecAck;

begin
   Int_32_Random.Reset(G);

end STCP;


